{- Administration interface.
   Allows to:
     - add, remove, search users
     - TODO: validate users
     - TODO: change user password
     - TODO: show user details (list of owned domains)
     - TODO: show user domain details (zone content) and to modify users' zone
     - TODO: raise a user to admin (and vice versa)
     - TODO: list users (getting them slowly, otherwise it will cause problems with thousands of logins)
-}
module App.AdministrationInterface where

import Prelude (Unit, bind, discard, not, pure, show, ($), (<<<), (<>), (=<<), map, (/=))

import Bulma as Bulma

import Data.Maybe (Maybe(..))
import Data.Array as A
import Effect.Aff.Class (class MonadAff)
import Halogen as H
import Halogen.HTML as HH
import Halogen.HTML.Events as HE
import Halogen.HTML.Properties as HP
import Web.Event.Event (Event)
import Web.Event.Event as Event

import CSSClasses as C

import Web.HTML (window) as HTML
import Web.HTML.Window (sessionStorage) as Window
import Web.Storage.Storage as Storage

import App.UserPublic (UserPublic)
import Data.ArrayBuffer.Types (ArrayBuffer)

import App.LogMessage
-- import App.IPC as IPC
import App.Email as Email

-- import App.Messages.DNSManagerDaemon as DNSManager
import App.Messages.AuthenticationDaemon as AuthD

data Output
  = MessageToSend ArrayBuffer
  | Log LogMessage
  | AskState
  | StoreState State

data Query a
  = MessageReceived AuthD.AnswerMessage a
  | ConnectionIsDown a
  | ConnectionIsUp a
  | ProvideState (Maybe State) a

type Slot = H.Slot Query Output

type Input = Unit

data AddUserInput
  = ADDUSER_INP_login       String
  | ADDUSER_INP_email       String
  | ADDUSER_toggle_admin
  | ADDUSER_INP_pass        String
  | SEARCHUSER_INP_regex    String
  --| SEARCHUSER_INP_domain String

data Action
  = HandleAddUserInput AddUserInput

  | AddUserAttempt
  | SearchUserAttempt
  | PreventSubmit Event

  | ShowUser Int
  | RemoveUser Int

  -- | Change the displayed page.
  | Routing Page

  | Initialize
  | Finalize

-- | There are different `sub-pages` in the authentication daemon admin page.
-- | For example, users can be searched and a list is provided.
data Page = Home | Search | Add

type StateAddUserForm = { login :: String, admin :: Boolean, email :: String, pass :: String }
type StateSearchUserForm = { regex :: String {-, admin :: Boolean, domain :: String -} }

type State =
  { addUserForm    :: StateAddUserForm
  , searchUserForm :: StateSearchUserForm
  , page           :: Page
  , wsUp           :: Boolean
  , matching_users :: Array UserPublic
  }

component :: forall m. MonadAff m => H.Component Query Input Output m
component =
  H.mkComponent
    { initialState
    , render
    , eval: H.mkEval $ H.defaultEval
        { initialize   = Just Initialize
        , handleAction = handleAction
        , handleQuery = handleQuery
        , finalize = Just Finalize
        }
    }

initialState :: Input -> State
initialState _ = { addUserForm:    { login: "", admin: false, email: "", pass: "" }
                 , searchUserForm: { regex: "" {-, admin: false, domain: "" -} }
                 , matching_users: []
                 , page: Home
                 , wsUp: true
                 }

render :: forall m. State -> H.ComponentHTML Action () m
render { addUserForm, searchUserForm, matching_users, page, wsUp }
    = HH.div_ 
        [ Bulma.box [routing_search_button, routing_add_button]
        , case page of
            Home -> Bulma.section_small [Bulma.h3 "Select an action"]
            Search -> Bulma.columns_
              [ Bulma.column (C.is 3) [Bulma.article (Bulma.p "Search users") render_searchuser_form]
              , Bulma.column_ [ Bulma.h3 "Result", show_found_users ]
              ]
            Add -> Bulma.columns_
              [ Bulma.column (C.is 5) [Bulma.article (Bulma.p "Add a new user") render_adduser_form] ]
        ]
    where
      show_found_users = Bulma.box [ HH.ul_ $ map user_card matching_users ]
      user_card user = HH.li_ [ Bulma.btn user.login (ShowUser user.uid)
                              , Bulma.alert_btn "remove" (RemoveUser user.uid)
                              ]
      up x = HandleAddUserInput <<< x
      active = (if wsUp then (HP.enabled true) else (HP.disabled true))

      render_adduser_form =
        HH.form
          [ HE.onSubmit PreventSubmit ]
          [ Bulma.box_input "login" "User login" "login" (up ADDUSER_INP_login) addUserForm.login active
          , Bulma.btn_labeled "adminbtn" "Admin" (show addUserForm.admin) (HandleAddUserInput ADDUSER_toggle_admin)
          , Bulma.box_input "email" "User email" "email" (up ADDUSER_INP_email) addUserForm.email active
          , Bulma.box_password "password" "User password" "password" (up ADDUSER_INP_pass) addUserForm.pass active
          , Bulma.btn "Send" AddUserAttempt
          ]

      render_searchuser_form =
        HH.form
          [ HE.onSubmit PreventSubmit ]
          [ Bulma.p """
                    Following input accepts any regex.
                    This will be used to search an user based on his login, full name or email address.
                    """
          , Bulma.box_input "regex" "Regex" "regex" (up SEARCHUSER_INP_regex) searchUserForm.regex active
          --, Bulma.btn_labeled "adminbtn" "Admin" (show searchUserForm.admin)
          --    (HandleAddUserInput SEARCHUSER_toggle_admin)
          --, Bulma.box_input "domain" "Domain owned" "blah.netlib.re."
          --    (up SEARCHUSER_INP_domain) searchUserForm.domain active
          , Bulma.btn "Send" SearchUserAttempt
          ]

      routing_search_button = Bulma.btn "Search" $ Routing Search
      routing_add_button = Bulma.btn "Add" $ Routing Add

handleAction :: forall m. MonadAff m => Action -> H.HalogenM State Action () Output m Unit
handleAction = case _ of
  Initialize -> do
    H.raise $ AskState
    sessionstorage <- H.liftEffect $ Window.sessionStorage =<< HTML.window
    old_page <- H.liftEffect $ Storage.getItem "current-ada-page" sessionstorage
    case old_page of
      Nothing -> H.raise $ Log $ ErrorLog "We hadn't changed page before reload apparently."
      Just page -> case page of
        "Home"    -> handleAction $ Routing Home
        "Search"  -> handleAction $ Routing Search
        "Add"     -> handleAction $ Routing Add
        _         -> H.raise $ Log $ ErrorLog $ "Reload but cannot understand old page: " <> page

  Finalize -> do
    state <- H.get
    H.raise $ StoreState state

  HandleAddUserInput adduserinp -> do
    { addUserForm } <- H.get
    case adduserinp of
      ADDUSER_INP_login  v -> H.modify_ _ { addUserForm { login     = v } }
      ADDUSER_INP_email  v -> H.modify_ _ { addUserForm { email     = v } }
      ADDUSER_toggle_admin -> H.modify_ _ { addUserForm { admin = not addUserForm.admin } }
      ADDUSER_INP_pass   v -> H.modify_ _ { addUserForm { pass      = v } }
      SEARCHUSER_INP_regex  v -> H.modify_ _ { searchUserForm { regex = v } }

  PreventSubmit ev -> H.liftEffect $ Event.preventDefault ev

  ShowUser uid -> do
    H.raise $ Log $ SystemLog $ "Show a user details (uid: " <> show uid <> ")"

  RemoveUser uid -> do
    H.raise $ Log $ SystemLog $ "Remove user " <> show uid
    ab <- H.liftEffect $ AuthD.serialize $ AuthD.MkDeleteUser { user: Just uid }
    H.raise $ MessageToSend ab

  AddUserAttempt -> do
    { addUserForm } <- H.get
    let login  = addUserForm.login
        email  = addUserForm.email
        pass   = addUserForm.pass

    case login, email, pass of
      "", _, _ -> H.raise $ Log $ UnableToSend "Write the user's login!"
      _, "", _ -> H.raise $ Log $ UnableToSend "Write the user's email!"
      _, _, "" -> H.raise $ Log $ UnableToSend "Write the user's password!"

      _, _, _ -> do
        ab <- H.liftEffect $ AuthD.serialize $
          AuthD.MkAddUser { login: login
                          , admin: addUserForm.admin
                          , email: Just (Email.Email email)
                          , password: pass }
        H.raise $ MessageToSend ab
        H.raise $ Log $ SystemLog "Add a user"

  Routing page -> do
    -- Store the current page we are on and restore it when we reload.
    sessionstorage <- H.liftEffect $ Window.sessionStorage =<< HTML.window
    _ <- case page of
      Home           -> H.liftEffect $ Storage.setItem "current-ada-page" "Home"   sessionstorage
      Search         -> H.liftEffect $ Storage.setItem "current-ada-page" "Search" sessionstorage
      Add            -> H.liftEffect $ Storage.setItem "current-ada-page" "Add"    sessionstorage
    H.modify_ _ { page = page }

  SearchUserAttempt -> do
    { searchUserForm } <- H.get
    let regex  = searchUserForm.regex
        -- domain = searchUserForm.domain
        -- admin  = searchUserForm.admin
    ab <- H.liftEffect $ AuthD.serialize $
      AuthD.MkSearchUser { regex: not_empty_string regex, offset: Just 0 }
    H.raise $ MessageToSend ab
    H.modify_ _ { matching_users = [] }

not_empty_string :: String -> Maybe String
not_empty_string "" = Nothing
not_empty_string v = Just v

handleQuery :: forall a m. MonadAff m => Query a -> H.HalogenM State Action () Output m (Maybe a)
handleQuery = case _ of

  ProvideState maybe_state a -> do
    case maybe_state of
      Nothing -> pure Nothing
      Just s  -> do
        H.put s
        pure (Just a)

  MessageReceived message a -> do
    case message of
      (AuthD.GotUserAdded msg) -> do
        H.raise $ Log $ SuccessLog $ "Added user: " <> show msg.user

      (AuthD.GotMatchingUsers msg) -> do
        H.raise $ Log $ SuccessLog "Got list of matched users."
        H.modify_ _ { matching_users = msg.users }

      (AuthD.GotUserDeleted msg) -> do
        H.raise $ Log $ SuccessLog $ "User (uid: " <> show msg.uid <> ") got removed."
        { matching_users } <- H.get
        H.modify_ _ { matching_users = A.filter (\x -> x.uid /= msg.uid) matching_users }

      -- Unexpected message.
      _ -> do
        H.raise $ Log $ ErrorLog $ "Authentication server didn't send a valid message."
    pure (Just a)

  ConnectionIsDown a -> do
    H.modify_ _ { wsUp = false }
    pure (Just a)

  ConnectionIsUp a -> do
    H.modify_ _ { wsUp = true }
    pure (Just a)
