module App.Validation.DNS where

import Prelude (apply, between, bind, map, pure, ($), (-), (<))

import Control.Alt ((<|>))
import Data.Array as A
import Data.Either (Either(..))
import Data.Maybe (Maybe(..), maybe)
import Data.String.CodeUnits as CU
import Data.Validation.Semigroup (V, invalid, toEither)

import App.ResourceRecord (ResourceRecord)
import GenericParser.SomeParsers as SomeParsers
import GenericParser.Parser as G
import GenericParser.DomainParser.Common (DomainError) as DomainParser
import GenericParser.DomainParser (sub_eof) as DomainParser
import GenericParser.IPAddress as IPAddress
import GenericParser.RFC5234 as RFC5234

-- | **History:**
-- | The module once used dedicated types for each type of RR.
-- | That comes with several advantages.
-- | First, type verification was a thing, and function were dedicated to a certain type of record.
-- | Second, these dedicated types used strings for their fields,
-- | which simplifies the typing when dealing with forms.
-- | Finally, the validation was a way to convert dedicated types (used in forms)
-- | to the general type (used for network serialization).
-- | This ensures each resource record is verified before being sent to `dnsmanagerd`.
-- |
-- | The problem is that, with dedicated types, you are then required to have dedicated functions.
-- | Conversion functions are also required.
-- |
-- | Maybe the code will change again in the future, but for now it will be enough.

data Error
  = UNKNOWN
  | VEIPv4 (G.Error IPAddress.IPv4Error)
  | VEIPv6 (G.Error IPAddress.IPv6Error)
  | VEName (G.Error DomainParser.DomainError)
  | VETTL Int Int Int
  | VETXT (G.Error TXTError)
  | VECNAME (G.Error DomainParser.DomainError)
  | VENS (G.Error DomainParser.DomainError)
  | VEMX (G.Error DomainParser.DomainError)
  | VEPriority Int Int Int
  | VESRV (G.Error DomainParser.DomainError)
  | VEProtocol (G.Error ProtocolError)
  | VEPort Int Int Int
  | VEWeight Int Int Int

type AVErrors = Array Error

-- | Current default values.
min_ttl :: Int
min_ttl = 30
max_ttl :: Int
max_ttl = 86000
max_txt :: Int
max_txt = 500
min_priority :: Int
min_priority = 0
max_priority :: Int
max_priority = 65535
min_port :: Int
min_port = 0
max_port :: Int
max_port = 65535
min_weight :: Int
min_weight = 0
max_weight :: Int
max_weight = 65535

-- Functions handling network-related structures (ResourceRecord).

type RRPriority = Maybe Int
type RRPort     = Maybe Int
type RRProtocol = Maybe String
type RRWeight   = Maybe Int
type RRMname    = Maybe String
type RRRname    = Maybe String
type RRSerial   = Maybe Int
type RRRefresh  = Maybe Int
type RRRetry    = Maybe Int
type RRExpire   = Maybe Int
type RRMinttl   = Maybe Int


toRR :: Int -> Boolean -> String -> String -> Int -> String
     -> RRPriority
     -> RRPort
     -> RRProtocol
     -> RRWeight
     -> RRMname
     -> RRRname
     -> RRSerial
     -> RRRefresh
     -> RRRetry
     -> RRExpire
     -> RRMinttl
     -> ResourceRecord
toRR rrid readonly rrtype rrname ttl target
     priority port protocol weight mname rname serial refresh retry expire minttl
  = { rrid:         rrid
    , readonly:     readonly
    , rrtype:       rrtype
    , name:         rrname
    , ttl:          ttl
    , target:       target

    -- MX + SRV
    , priority:     priority

    -- SRV
    , port:         port
    , protocol:     protocol
    , weight:       weight

    -- SOA
    , mname:        mname
    , rname:        rname
    , serial:       serial
    , refresh:      refresh
    , retry:        retry
    , expire:       expire
    , minttl:       minttl
    }

toRR_basic :: Int -> Boolean -> String -> String -> Int -> String -> ResourceRecord
toRR_basic rrid readonly rrtype rrname ttl target
  = toRR rrid readonly rrtype rrname ttl target
         Nothing Nothing Nothing Nothing Nothing Nothing Nothing Nothing Nothing Nothing Nothing

-- last + priority
toRR_mx :: Int -> Boolean -> String -> String -> Int -> String -> Int -> ResourceRecord
toRR_mx rrid readonly rrtype rrname ttl target priority
  = toRR rrid readonly rrtype rrname ttl target (Just priority)
         Nothing Nothing Nothing Nothing Nothing Nothing Nothing Nothing Nothing Nothing

-- last + port + protocol + weight
toRR_srv :: Int -> Boolean -> String -> String -> Int -> String -> Int -> Int -> String -> Int -> ResourceRecord
toRR_srv rrid readonly rrtype rrname ttl target priority port protocol weight
  = toRR rrid readonly rrtype rrname ttl target (Just priority) (Just port) (Just protocol) (Just weight)
         Nothing Nothing Nothing Nothing Nothing Nothing Nothing

data TXTError
  = TXTInvalidCharacter
  | TXTTooLong Int Int -- max current
-- | TODO: `txt_parser` is currently accepting any printable character (`vchar + sp`).
txt_parser :: G.Parser TXTError String
txt_parser = do pos <- G.current_position
                v <- A.many (RFC5234.vchar <|> RFC5234.sp)
                e <- G.tryMaybe SomeParsers.eof
                pos2 <- G.current_position
                case e of
                  Nothing -> G.errorParser $ Just TXTInvalidCharacter
                  Just _ -> do
                    let nbchar = pos2 - pos
                    if nbchar < max_txt
                    then pure $ CU.fromCharArray v
                    else G.Parser \_ -> G.failureError pos (Just $ TXTTooLong max_txt nbchar)

-- | `parse` allows to run any parser based on `GenericParser` and provide a validation error.
-- | The actual validation error contains the parser's error including the position.
parse :: forall e v. G.Parser e v -> String -> ((G.Error e) -> Error) -> V (Array Error) v
parse (G.Parser p) str c = case p { string: str, position: 0 } of
  Left x  -> invalid $ [c x]
  Right x -> pure x.result

validationA :: ResourceRecord -> V (Array Error) ResourceRecord
validationA form = ado
  name   <- parse DomainParser.sub_eof form.name   VEName
  ttl    <- is_between min_ttl max_ttl form.ttl    VETTL
  target <- parse IPAddress.ipv4       form.target VEIPv4
  in toRR_basic form.rrid form.readonly "A" name ttl target

validationAAAA :: ResourceRecord -> V (Array Error) ResourceRecord
validationAAAA form = ado
  name   <- parse DomainParser.sub_eof form.name VEName
  ttl    <- is_between min_ttl max_ttl form.ttl  VETTL
  -- use read_input to get unaltered input (the IPv6 parser expands the input)
  target <- parse (G.read_input IPAddress.ipv6) form.target VEIPv6
  in toRR_basic form.rrid form.readonly "AAAA" name ttl target

validationTXT :: ResourceRecord -> V (Array Error) ResourceRecord
validationTXT form = ado
  name   <- parse DomainParser.sub_eof form.name   VEName
  ttl    <- is_between min_ttl max_ttl form.ttl    VETTL
  target <- parse txt_parser           form.target VETXT
  in toRR_basic form.rrid form.readonly "TXT" name ttl target

validationCNAME :: ResourceRecord -> V (Array Error) ResourceRecord
validationCNAME form = ado
  name   <- parse DomainParser.sub_eof form.name VEName
  ttl    <- is_between min_ttl max_ttl form.ttl      VETTL
  target <- parse DomainParser.sub_eof form.target VECNAME
  in toRR_basic form.rrid form.readonly "CNAME" name ttl target

validationNS :: ResourceRecord -> V (Array Error) ResourceRecord
validationNS form = ado
  name   <- parse DomainParser.sub_eof form.name   VEName
  ttl    <- is_between min_ttl max_ttl form.ttl    VETTL
  target <- parse DomainParser.sub_eof form.target VENS
  in toRR_basic form.rrid form.readonly "NS" name ttl target

data ProtocolError
  = InvalidProtocol

protocol_parser :: G.Parser ProtocolError String
protocol_parser = do
  G.string "tcp" <|> G.string "udp" G.<:> \_ -> InvalidProtocol

is_between :: Int -> Int -> Int -> (Int -> Int -> Int -> Error) -> V (Array Error) Int
is_between min max n ve = if between min max n
                          then pure n
                          else invalid [ve min max n]

validationMX :: ResourceRecord -> V (Array Error) ResourceRecord
validationMX form = ado
  name     <- parse DomainParser.sub_eof form.name     VEName
  ttl      <- is_between min_ttl max_ttl form.ttl      VETTL
  target   <- parse DomainParser.sub_eof form.target   VEMX
  priority <- is_between min_priority max_priority (maybe 0 id form.priority) VEPriority
  in toRR_mx form.rrid form.readonly "MX" name ttl target priority

validationSRV :: ResourceRecord -> V (Array Error) ResourceRecord
validationSRV form = ado
  name     <- parse DomainParser.sub_eof form.name     VEName
  ttl      <- is_between min_ttl max_ttl form.ttl      VETTL
  target   <- parse DomainParser.sub_eof form.target   VESRV
  priority <- is_between min_priority max_priority (maybe 0 id form.priority) VEPriority
  protocol <- parse protocol_parser      (maybe "" id   form.protocol)  VEProtocol
  port     <- is_between min_port max_port (maybe 0 id form.port)       VEPort
  weight   <- is_between min_weight max_weight (maybe 0 id form.weight) VEWeight
  in toRR_srv form.rrid form.readonly "SRV" name ttl target priority port protocol weight

validation :: ResourceRecord -> Either (Array Error) ResourceRecord
validation entry = case entry.rrtype of
  "A"     -> toEither $ validationA     entry
  "AAAA"  -> toEither $ validationAAAA  entry
  "TXT"   -> toEither $ validationTXT   entry
  "CNAME" -> toEither $ validationCNAME entry
  "NS"    -> toEither $ validationNS    entry
  "MX"    -> toEither $ validationMX    entry
  "SRV"   -> toEither $ validationSRV   entry
  _       -> toEither $ invalid [UNKNOWN]

id :: forall a. a -> a
id x = x
