-- | `App.ZoneInterface` provides an interface to display and modify a DNS zone.
-- |
-- | This interface allows to:
-- |   - display all resource records of a zone (SOA, NS, A, AAAA, CNAME, TXT, MX, SRV)
-- |   - add, modify, remove resource records
-- |
-- | **WIP**: Display relevant information for each record type in the (add/mod) modal.
-- | This includes explaining use cases and displaying an appropriate interface for the
-- | task at hand. For example, having a dedicated interface for DKIM.
-- |
-- | TODO: display errors not only for a record but for the whole zone.
-- | A DNS zone is bound by a set of rules, the whole zone must be consistent.
-- | For example, a CNAME `target` has to point to the `name` of an existing record.
-- |
-- | TODO: do not allow for the modification of read-only resource records.
-- |
-- | TODO: move all serialization code to a single module.
module App.ZoneInterface where

import Prelude (Unit, unit, void
               , bind, pure
               , not, comparing, discard, map, show
               , (&&), ($), (/=), (<<<), (<>), (==), (>), (#))

import Data.Array as A
import Data.Int (fromString)
import Data.ArrayBuffer.Types (ArrayBuffer)
import Data.Array.NonEmpty as NonEmpty
import Data.Either (Either(..))
-- import Data.Foldable as Foldable
import Data.Maybe (Maybe(..), fromMaybe, maybe)
import Effect.Aff.Class (class MonadAff)
import Halogen as H
import Halogen.HTML as HH
import Halogen.HTML.Properties as HP

import Bulma as Bulma
import CSSClasses as C

import App.AcceptedRRTypes (AcceptedRRTypes(..))
import App.ResourceRecord (ResourceRecord)

import App.DisplayErrors (error_to_paragraph)

import App.LogMessage (LogMessage(..))
import App.Messages.DNSManagerDaemon as DNSManager
import App.Validation.DNS as Validation

type RRId = Int

id :: forall a. a -> a
id x = x

-- | `App.ZoneInterface` can send messages through websocket interface
-- | connected to dnsmanagerd. See `App.WS`.
-- |
-- | Also, this component can log messages and ask its parent (`App.Container`) to
-- | reconnect the websocket to `dnsmanagerd`.

data Output
  = MessageToSend ArrayBuffer
  | Log LogMessage

-- | `App.ZoneInterface` can receive messages from `dnsmanagerd`.
-- |
-- | The component is also informed when the connection is lost or up again.

data Query a
  = MessageReceived DNSManager.AnswerMessage a
  | ConnectionIsDown a
  | ConnectionIsUp a

type Slot = H.Slot Query Output

-- | `App.ZoneInterface` has a single input: the domain name.

type Input = String

data Field
  = Field_Domain    String
  | Field_TTL       String
  | Field_Target    String
  | Field_Priority  String
  | Field_Protocol  String
  | Field_Weight    String
  | Field_Port      String

-- | Steps to create a new RR:
-- | 1. `CreateNewRRModal AcceptedRRTypes`: create a modal with default values based on selected accepted type.
-- | 2. `UpdateCurrentRR Field`: modify the fields of the future new RR.
-- | 3. `ValidateRR AcceptedRRTypes`: validate the new RR stored in `_currentRR`.
-- |    In case it works, automatically call `AddRR` then `CancelModal`.
-- | 4. `AddRR AcceptedRRTypes ResourceRecord`: send a message to `dnsmanagerd`.
-- |
-- | Steps to update an entry:
-- | 1. `CreateUpdateRRModal RRId`: create a modal from the values of the RR in `_resources` to update.
-- | 2. `UpdateCurrentRR Field`: modify the currently displayed RR.
-- | 3. `ValidateLocal RRId AcceptedRRTypes`: validate the RR.
-- | 4. `SaveRR ResourceRecord`: save the _validated_ RR by sending a message to `dnsmanagerd`.

data Action
  -- | Initiate the component. This means asking the content of the zone to `dnsmanagerd`.
  = Initialize

  -- | Cancel the current displayed modal.
  | CancelModal

  -- | Create a new resource record modal (a form) for a certain type of component.
  | CreateNewRRModal AcceptedRRTypes

  -- | Create modal (a form) for a resource record to update.
  | CreateUpdateRRModal RRId

  -- | Create a modal to ask confirmation before deleting a resource record.
  | DeleteRRModal RRId

  -- | Update new entry form (in the `rr_modal` modal).
  | UpdateCurrentRR Field

  -- | Validate a new resource record before adding it.
  | ValidateRR AcceptedRRTypes

  -- | Validate the entries in an already existing resource record.
  -- | Automatically calls for `SaveRR` once record is verified.
  | ValidateLocal

  -- | Add a new resource record to the zone.
  | AddRR AcceptedRRTypes ResourceRecord

  -- | Save the changes done in an already existing resource record.
  | SaveRR ResourceRecord

  -- | Send a message to remove a resource record.
  -- | Automatically closes the modal.
  | RemoveRR RRId

  -- | Ask `dnsmanagerd` for the generated zone file.
  | AskZoneFile

data RRModal
  = NoModal
  | NewRRModal AcceptedRRTypes
  | UpdateRRModal
  | RemoveRRModal RRId

show_accepted_type :: AcceptedRRTypes -> String
show_accepted_type = case _ of
  A     -> "A"
  AAAA  -> "AAAA"
  TXT   -> "TXT"
  CNAME -> "CNAME"
  NS    -> "NS"
  MX    -> "MX"
  SRV   -> "SRV"

string_to_acceptedtype :: String -> Maybe AcceptedRRTypes
string_to_acceptedtype str = case str of
  "A"     -> Just     A
  "AAAA"  -> Just  AAAA
  "TXT"   -> Just   TXT
  "CNAME" -> Just CNAME
  "NS"    -> Just    NS
  "MX"    -> Just    MX
  "SRV"   -> Just   SRV
  _       -> Nothing

type State =
  { _domain :: String
  , wsUp :: Boolean

  -- A modal to present a form for adding a new RR.
  , rr_modal :: RRModal

  -- | All resource records.
  , _resources :: Array ResourceRecord
  --, _local_errors :: Hash.HashMap RRId (Array Validation.Error)

  -- Unique RR form.
  , _currentRR :: ResourceRecord
  , _currentRR_errors :: Array Validation.Error

  , _zonefile :: Maybe String
  }

component :: forall m. MonadAff m => H.Component Query Input Output m
component =
  H.mkComponent
    { initialState
    , render
    , eval: H.mkEval $ H.defaultEval
        { initialize   = Just Initialize
        , handleAction = handleAction
        , handleQuery  = handleQuery
        }
    }

-- | Default available domain: netlib.re.

default_domain :: String
default_domain = "netlib.re"

default_empty_rr :: ResourceRecord
default_empty_rr
  = { rrtype: "A"
    , rrid: 0
    , name: "www"
    , ttl: 1800
    , target: "10.0.0.1"
    , readonly: false

    -- MX (and SRV) specific entry.
    , priority: Nothing

    -- SRV specific entries.
    , port: Nothing
    , protocol: Nothing
    , weight: Nothing

    -- SOA specific entries.
    , mname: Nothing
    , rname: Nothing
    , serial: Nothing
    , refresh: Nothing
    , retry: Nothing
    , expire: Nothing
    , minttl: Nothing
    }

initialState :: Input -> State
initialState domain =
  { wsUp: true
  , rr_modal: NoModal

  , _domain: domain

  , _resources: []
  --, _local_errors: Hash.empty

  -- This is the state for the new RR modal.
  , _currentRR: default_empty_rr
  -- List of errors within the form in new RR modal.
  , _currentRR_errors: []
  , _zonefile: Nothing
  }

type SortableRecord l = Record (rrtype :: String, rrid :: Int | l)

render :: forall m. State -> H.ComponentHTML Action () m
render state
  = Bulma.section_small
      [ case state.wsUp, state.rr_modal of
          false, _                   -> Bulma.p "You are disconnected."
          true,  RemoveRRModal rr_id -> modal_rr_delete rr_id
          true,  NewRRModal _        -> render_current_rr_modal
          true,  UpdateRRModal       -> render_current_rr_modal
          true,  NoModal -> HH.div_
            [ Bulma.h1 state._domain
            , Bulma.hr
            , render_resources $ sorted state._resources
            , Bulma.hr
            , render_new_records state
            , render_zonefile state._zonefile
            ]
      ]
  where
    sorted :: forall l. Array (SortableRecord (l)) -> Array (SortableRecord (l))
    sorted array =
      A.groupAllBy (comparing (_.rrtype)) array    -- [x2 z2 x1 y z1 z3] -> [ NE[x2x1] NE[y] NE[z2z1z3] ]
      # map NonEmpty.toArray                       -- -> [[x2x1][y][z2z1z3]]
      # map (A.sortBy (comparing (_.rrid)))        -- -> [[x1x2][y][z1z2z3]]
      # A.concat                                   -- -> [x1 x2 y z1 z2 z3]

    modal_rr_delete :: forall w. Int -> HH.HTML w Action
    modal_rr_delete rr_id = Bulma.modal "Deleting a resource record"
      [warning_message] [modal_delete_button, Bulma.cancel_button CancelModal]
      where
      modal_delete_button = Bulma.alert_btn "Delete the resource record." (RemoveRR rr_id)
      warning_message
        = HH.p [] [ HH.text "You are about to delete a resource record, this actions is "
                  , Bulma.strong "irreversible"
                  , HH.text "."
                  ]

    render_current_rr_modal :: forall w. HH.HTML w Action
    render_current_rr_modal =
      case state._currentRR.rrtype of
      "A"     -> template content_simple (foot_content A)
      "AAAA"  -> template content_simple (foot_content AAAA)
      "TXT"   -> template content_simple (foot_content TXT)
      "CNAME" -> template content_simple (foot_content CNAME)
      "NS"    -> template content_simple (foot_content NS)
      "MX"    -> template content_mx     (foot_content MX)
      "SRV"   -> template content_srv    (foot_content SRV)
      _       -> Bulma.p $ "Invalid type: " <> state._currentRR.rrtype
      where
        -- DRY
        updateForm x = UpdateCurrentRR <<< x
        render_errors = if A.length state._currentRR_errors > 0
                        then HH.div_ $ [ Bulma.h3 "Errors: " ] <> map error_to_paragraph state._currentRR_errors
                        else HH.div_ [ ]
        content_simple :: Array (HH.HTML w Action)
        content_simple =
          [ render_errors
          --, Bulma.box_input ("domain" <> state._currentRR.rrtype) "Name" "www"   -- id, title, placeholder
          --    (updateForm Field_Domain)             -- action
          --    state._currentRR.name                            -- value
          --    should_be_disabled                           -- condition
          , Bulma.input_with_side_text ("domain" <> state._currentRR.rrtype) "Name" "www" -- id, title, placeholder
              (updateForm Field_Domain)             -- action
              state._currentRR.name                 -- value
              ("." <> state._domain)                -- sidetext
          , Bulma.box_input ("ttl" <> state._currentRR.rrtype) "TTL" "600"
              (updateForm Field_TTL)
              (show state._currentRR.ttl)
              should_be_disabled
          , Bulma.box_input ("target" <> state._currentRR.rrtype) "Target" "198.51.100.5"
              (updateForm Field_Target)
              state._currentRR.target
              should_be_disabled
          ]
        content_mx :: Array (HH.HTML w Action)
        content_mx =
          [ render_errors
          , Bulma.input_with_side_text "domainMX" "Name" "www" -- id, title, placeholder
              (updateForm Field_Domain)             -- action
              state._currentRR.name                 -- value
              ("." <> state._domain)                -- sidetext
          , Bulma.box_input ("ttlMX") "TTL" "600"
              (updateForm Field_TTL)
              (show state._currentRR.ttl)
              should_be_disabled
          , Bulma.box_input ("targetMX") "Target" "www"
              (updateForm Field_Target)
              state._currentRR.target
              should_be_disabled
          , Bulma.box_input ("priorityMX") "Priority" "10"
              (updateForm Field_Priority)
              (maybe "" show state._currentRR.priority)
              should_be_disabled
          ]
        content_srv :: Array (HH.HTML w Action)
        content_srv =
          [ render_errors
          , Bulma.input_with_side_text "domainSRV" "Name" "www" -- id, title, placeholder
              (updateForm Field_Domain)             -- action
              state._currentRR.name                 -- value
              ("." <> state._domain)                -- sidetext
          , Bulma.box_input ("ttlSRV") "TTL" "600"
              (updateForm Field_TTL)
              (show state._currentRR.ttl)
              should_be_disabled
          , Bulma.box_input ("targetSRV") "Target" "www"
              (updateForm Field_Target)
              state._currentRR.target
              should_be_disabled
          , Bulma.box_input ("prioritySRV") "Priority" "10"
              (updateForm Field_Priority)
              (maybe "" show state._currentRR.priority)
              should_be_disabled
          , Bulma.box_input ("portSRV") "Port" "5061"
              (updateForm Field_Port)
              (maybe "" show state._currentRR.port)
              should_be_disabled
          , Bulma.box_input ("weightSRV") "Weight" "100"
              (updateForm Field_Weight)
              (maybe "" show state._currentRR.weight)
              should_be_disabled
          , Bulma.box_input ("protocolSRV") "Protocol" "tcp"
              (updateForm Field_Protocol)
              (fromMaybe "tcp" state._currentRR.protocol)
              should_be_disabled
          ]

        should_be_disabled = (if true then (HP.enabled true) else (HP.disabled true))
        foot_content x = [ case state.rr_modal of
                             NewRRModal _ -> Bulma.btn_add (ValidateRR x)
                             UpdateRRModal -> Bulma.btn_save ValidateLocal
                             _ -> Bulma.p "state.rr_modal should either be NewRRModal or UpdateRRModal."
                         ]
        template content foot_ = Bulma.modal title content foot
          where
          title = case state.rr_modal of
            NoModal             -> "Error: no modal should be displayed"
            NewRRModal t_       -> "New " <> show_accepted_type t_ <> " resource record"
            UpdateRRModal       -> "Update RR " <> show state._currentRR.rrid <> " resource record"
            RemoveRRModal rr_id -> "Error: should display removal modal instead (for RR " <> show rr_id <> ")"
          foot = foot_ <> [Bulma.cancel_button CancelModal]

handleAction :: forall m. MonadAff m => Action -> H.HalogenM State Action () Output m Unit
handleAction = case _ of
  -- | Cancel the current modal being presented.
  -- | Works for both "new RR", "update RR" and "remove RR" modals.
  CancelModal -> do
    H.modify_ _ { rr_modal = NoModal }
    H.modify_ _ { _currentRR_errors = [] }

  -- | Create the RR modal.
  DeleteRRModal rr_id -> do
    H.modify_ _ { rr_modal = RemoveRRModal rr_id }

  -- | Create modal (a form) for a resource record to update.
  CreateUpdateRRModal rr_id -> do
    state <- H.get
    case first (\rr -> rr.rrid == rr_id) state._resources of
      Nothing -> H.raise $ Log $ ErrorLog $ "RR not found (RR " <> show rr_id <> ")"
      Just rr -> do
        H.modify_ _ { _currentRR = rr }
        H.modify_ _ { rr_modal = UpdateRRModal }

  -- | Each time a "new RR" button is clicked, the form resets.
  CreateNewRRModal t -> do
    state <- H.get
    H.modify_ _ { rr_modal = NewRRModal t }
    let defaultA = { rrtype: "A"
          , rrid: 0
          , ttl: 600
          , readonly: false
          , name:     "www"
          , target:   "192.0.2.1"
          , port: Nothing
          , weight: Nothing
          , priority: Nothing
          , protocol: Nothing
          , mname: Nothing
          , rname: Nothing
          , serial: Nothing
          , refresh: Nothing
          , retry: Nothing
          , expire: Nothing
          , minttl: Nothing
          }
        defaultAAAA = { rrtype: "AAAA"
          , rrid: 0
          , ttl: 600
          , readonly: false
          , name:     "www"
          , target:   "2001:db8::1"
          , port: Nothing
          , weight: Nothing
          , priority: Nothing
          , protocol: Nothing
          , mname: Nothing
          , rname: Nothing
          , serial: Nothing
          , refresh: Nothing
          , retry: Nothing
          , expire: Nothing
          , minttl: Nothing
          }
        defaultTXT = { rrtype: "TXT"
          , rrid: 0
          , ttl: 600
          , readonly: false
          , name:     "txt"
          , target:   "some text"
          , port: Nothing
          , weight: Nothing
          , priority: Nothing
          , protocol: Nothing
          , mname: Nothing
          , rname: Nothing
          , serial: Nothing
          , refresh: Nothing
          , retry: Nothing
          , expire: Nothing
          , minttl: Nothing
          }
        defaultCNAME = { rrtype: "CNAME"
          , rrid: 0
          , ttl: 600
          , readonly: false
          , name:     "blog"
          , target:   "www"
          , port: Nothing
          , weight: Nothing
          , priority: Nothing
          , protocol: Nothing
          , mname: Nothing
          , rname: Nothing
          , serial: Nothing
          , refresh: Nothing
          , retry: Nothing
          , expire: Nothing
          , minttl: Nothing
          }
        defaultNS = { rrtype: "NS"
          , rrid: 0
          , ttl: 600
          , readonly: false
          , name:     (state._domain <> ".")
          , target:   "ns0.example.com."
          , port: Nothing
          , weight: Nothing
          , priority: Nothing
          , protocol: Nothing
          , mname: Nothing
          , rname: Nothing
          , serial: Nothing
          , refresh: Nothing
          , retry: Nothing
          , expire: Nothing
          , minttl: Nothing
          }
        defaultMX = { rrtype: "MX"
          , rrid: 0
          , ttl: 600
          , readonly: false
          , name:     "mail"
          , target:   "www"
          , port: Nothing
          , weight: Nothing
          , priority: Just 10
          , protocol: Nothing
          , mname: Nothing
          , rname: Nothing
          , serial: Nothing
          , refresh: Nothing
          , retry: Nothing
          , expire: Nothing
          , minttl: Nothing
          }
        defaultSRV = { rrtype: "SRV"
          , rrid: 0
          , ttl: 600
          , readonly: false
          , name:     "_sip._tcp"
          , target:   "www"
          , port: Just 5061
          , weight: Just 100
          , priority: Just 10
          , protocol: Just "tcp"
          , mname: Nothing
          , rname: Nothing
          , serial: Nothing
          , refresh: Nothing
          , retry: Nothing
          , expire: Nothing
          , minttl: Nothing
          }
    case t of
      A     -> H.modify_ _ { _currentRR = defaultA }
      AAAA  -> H.modify_ _ { _currentRR = defaultAAAA }
      TXT   -> H.modify_ _ { _currentRR = defaultTXT }
      CNAME -> H.modify_ _ { _currentRR = defaultCNAME }
      NS    -> H.modify_ _ { _currentRR = defaultNS }
      MX    -> H.modify_ _ { _currentRR = defaultMX }
      SRV   -> H.modify_ _ { _currentRR = defaultSRV }

  -- | Initialize the ZoneInterface component: ask for the domain zone to `dnsmanagerd`.
  Initialize -> do
    { _domain } <- H.get
    H.raise $ Log $ SystemLog $ "Asking the domain " <> _domain
    message <- H.liftEffect $ DNSManager.serialize $ DNSManager.MkGetZone { domain: _domain }
    H.raise $ MessageToSend message

  -- | Perform validation. In case the record is valid, it is added to the zone then the modal is closed.
  -- | Else, the different errors are added to the state.
  ValidateRR t -> do
    state <- H.get
    case Validation.validation state._currentRR of
      Left actual_errors -> do
        -- H.raise $ Log $ ErrorLog $ "Cannot add this " <> show t <> " RR, some errors occured in the record:"
        -- loopE (\v -> H.raise $ Log $ ErrorLog $ "==> " <> show_error v) actual_errors
        H.modify_ _ { _currentRR_errors = actual_errors }
      Right newrr -> do
        H.modify_ _ { _currentRR_errors = [] }
        handleAction $ AddRR t newrr
        handleAction CancelModal

  -- | Try to add a resource record to the zone.
  -- | Can fail if the content of the form isn't valid.
  AddRR t newrr -> do
    state <- H.get
    H.raise $ Log $ SystemLog $ "Add new " <> show t
    message <- H.liftEffect
      $ DNSManager.serialize
      $ DNSManager.MkAddRR { domain: state._domain, rr: newrr }
    H.raise $ MessageToSend message

  -- | Update the currently displayed RR form (new or update RR).
  UpdateCurrentRR field -> do
    state <- H.get
    let newRR = update_field state._currentRR field
    H.modify_ _ { _currentRR = newRR }

  -- | Validate any local RR with the new `_resources` and `_local_errors`.
  ValidateLocal -> do
    state <- H.get
    case Validation.validation state._currentRR of
      Left actual_errors -> do
        H.modify_ _ { _currentRR_errors = actual_errors }
      Right rr -> do
        H.modify_ _ { _currentRR_errors = [] }
        handleAction $ SaveRR rr

  SaveRR rr -> do
    state <- H.get
    H.raise $ Log $ SystemLog $ "Updating RR " <> show rr.rrid
    message <- H.liftEffect
      $ DNSManager.serialize
      $ DNSManager.MkUpdateRR { domain: state._domain, rr: rr }
    H.raise $ MessageToSend message

  RemoveRR rr_id -> do
    { _domain } <- H.get
    H.raise $ Log $ SystemLog $ "Ask to remove rr (rrid: " <> show rr_id <> ")"
    -- Send a removal message.
    message <- H.liftEffect
      $ DNSManager.serialize
      $ DNSManager.MkDeleteRR { domain: _domain, rrid: rr_id }
    H.raise $ MessageToSend message
    -- Modal doesn't need to be active anymore.
    handleAction CancelModal

  AskZoneFile -> do
    state <- H.get
    H.raise $ Log $ SystemLog $ "Asking for the '" <> state._domain <> "' zonefile"
    message <- H.liftEffect
      $ DNSManager.serialize
      $ DNSManager.MkAskGeneratedZoneFile { domain: state._domain }
    H.raise $ MessageToSend message

handleQuery :: forall a m. MonadAff m => Query a -> H.HalogenM State Action () Output m (Maybe a)
handleQuery = case _ of

  MessageReceived message a -> do
    case message of
      (DNSManager.MkRRUpdated response) -> do
        replace_entry response.rr
        -- When an update is received for a record, it means
        -- the update request has been accepted, the current modal can be closed.
        H.modify_ _ { rr_modal = NoModal }
      (DNSManager.MkRRAdded response) -> do
        state <- H.get
        H.put $ add_RR state response.rr
      (DNSManager.MkRRDeleted response) -> do
        -- Remove the resource record.
        state <- H.get
        H.modify_ _ { _resources = A.filter (\rr -> rr.rrid /= response.rrid) state._resources }
      (DNSManager.MkGeneratedZoneFile response) -> do
        H.modify_ _ { _zonefile = Just response.zonefile }
      (DNSManager.MkZone response) -> do
        add_entries response.zone.resources

      _ -> H.raise $ Log $ ErrorLog $ "Message not handled in ZoneInterface."
    pure (Just a)

  ConnectionIsDown a -> do
    H.modify_ _ { wsUp = false }
    pure (Just a)

  ConnectionIsUp a -> do
    H.modify_ _ { wsUp = true }
    pure (Just a)

  where
    -- replace_entry :: ResourceRecord
    replace_entry new_rr = do
      state <- H.get
      H.modify_ _ { _resources = A.filter (\rr -> rr.rrid /= new_rr.rrid) state._resources }

      new_state <- H.get
      H.put $ add_RR new_state new_rr

    add_entries []  = pure unit
    add_entries arr = do
      case A.head arr, A.tail arr of
        Nothing, _ -> pure unit
        Just new_rr, tail -> do
            state <- H.get
            H.put $ add_RR state new_rr
            add_entries $ fromMaybe [] tail

    add_RR :: State -> ResourceRecord -> State
    add_RR state new_rr = state { _resources = (state._resources <> [ new_rr ]) }

-- Rendering
render_soa :: forall (w :: Type). Maybe ResourceRecord -> HH.HTML w Action
render_soa Nothing    = Bulma.box [ HH.text "SOA not loaded, yet" ]
render_soa (Just soa) = Bulma.box [ Bulma.zone_rr_title "Start Of Authority (SOA)"
                                  , table_rr
                                  ]
  where table_rr = Bulma.table [] [ Bulma.soa_table_header, table_content ]
        table_content
          = HH.tbody_ $ [ HH.tr_ $ [
                                   --, Bulma.p $ "rrtype: "   <>      soa.rrtype
                                   --, Bulma.p $ "rrid: "     <> show soa.rrid
                                     HH.td_ [ HH.text soa.name ]
                                   , HH.td_ [ HH.text $ show soa.ttl ]
                                   , HH.td_ [ HH.text soa.target ]
                                   , HH.td_ [ HH.text $ maybe "" id soa.mname   ]
                                   , HH.td_ [ HH.text $ maybe "" id soa.rname   ]
                                   , HH.td_ [ HH.text $ maybe "" show soa.serial  ]
                                   , HH.td_ [ HH.text $ maybe "" show soa.refresh ]
                                   , HH.td_ [ HH.text $ maybe "" show soa.retry   ]
                                   , HH.td_ [ HH.text $ maybe "" show soa.expire  ]
                                   , HH.td_ [ HH.text $ maybe "" show soa.minttl  ]
                                   ]
                        ]

-- | Render all Resource Records.
render_resources :: forall w
  -- .  Hash.HashMap RRId (Array Validation.Error)
  . Array (ResourceRecord)
  -> HH.HTML w Action
render_resources []
  = Bulma.box [ Bulma.zone_rr_title "Resource records"
              , Bulma.subtitle "No records for now"
              ]
render_resources records
  = HH.div_ [ render_soa $ A.head $ A.filter (\rr -> rr.rrtype == "SOA") records
            , Bulma.box [ basic_records_section ]
            , Bulma.box [ mx_records_section    ]
            , Bulma.box [ srv_records_section   ]
            , Bulma.box_ C.has_background_warning_light [ basic_readonly_records_section ]
            ]
  where
  all_basic_rr    = A.filter (\rr -> A.elem rr.rrtype baseRecords && not rr.readonly) records
  all_basic_ro_rr = A.filter (\rr -> A.elem rr.rrtype baseRecords && rr.readonly)     records
  all_mx_rr    = A.filter (\rr -> rr.rrtype == "MX") records
  all_srv_rr   = A.filter (\rr -> rr.rrtype == "SRV") records

  basic_records_section
    = if A.length all_basic_rr > 0
      then Bulma.table [] [ Bulma.simple_table_header, render_basic_records all_basic_rr]
      else Bulma.p "no basic records"

  basic_readonly_records_section
    = if A.length all_basic_ro_rr > 0
      then Bulma.table [] [ Bulma.simple_table_header_ro, render_basic_records all_basic_ro_rr]
      else Bulma.p "no read only records?"

  mx_records_section
    = if A.length all_mx_rr > 0
      then Bulma.table [] [ Bulma.mx_table_header, render_mx_records ]
      else Bulma.p "no mx records"

  srv_records_section
    = if A.length all_srv_rr > 0
      then Bulma.table [] [ Bulma.srv_table_header, render_srv_records ]
      else Bulma.p "no srv records"

  render_basic_records _rrs = table_content_with_separations _rrs
  render_mx_records = table_content all_mx_rr
  render_srv_records = table_content all_srv_rr

  table_content_with_separations records_ = HH.tbody_ $
    A.groupAllBy (comparing (_.rrtype)) records_ -- [x x y y z] -> [NE[xx], NE[yy], NE[z]]
    # map NonEmpty.toArray                       -- -> [[xx], [yy], [z]]
    # map (map rows)                             -- -> [[hh], [hh], [h]] ('h' means 'html')
    # A.intersperse [emptyline]                  -- -> [[hh], [line], [hh], [line], [h]]
    # A.concat                                   -- -> [h h line h h line h]

  emptyline = HH.tr_ [ Bulma.txt_name "", HH.td_ [], HH.td_ [], HH.td_ [], HH.td_ [], HH.td_ [] ]

  table_content records_ = HH.tbody_ $ map rows records_
  rows rr = if rr.readonly
            then HH.tr [ HP.classes C.has_background_warning_light ] $ render_row rr -- <> error_row rr
            else HH.tr_ $ render_row rr -- <> error_row rr

  render_row :: ResourceRecord -> Array (HH.HTML w Action)
  render_row rr =
    case rr.rrtype of
      "SRV" ->
        [ Bulma.txt_name rr.rrtype
        , HH.td_ [ Bulma.p rr.name]
        , HH.td_ [ Bulma.p $ show rr.ttl ]
        , HH.td_ [ Bulma.p $ maybe "" show rr.priority ]
        , HH.td_ [ Bulma.p $ maybe "" id rr.protocol ]
        , HH.td_ [ Bulma.p $ maybe "" show rr.weight ]
        , HH.td_ [ Bulma.p $ maybe "" show rr.port ]
        , HH.td_ [ Bulma.p rr.target ]
        , if rr.readonly
          then HH.td_ [ Bulma.btn_readonly ]
          else HH.td_ [ Bulma.btn_modify (CreateUpdateRRModal rr.rrid) ]
        , if rr.readonly
          then HH.td_ [ Bulma.btn_readonly ]
          else HH.td_ [ Bulma.btn_delete (DeleteRRModal rr.rrid) ]
        ]
      "MX"  ->
        [ Bulma.txt_name rr.rrtype
        , HH.td_ [ Bulma.p rr.name]
        , HH.td_ [ Bulma.p $ show rr.ttl ]
        , HH.td_ [ Bulma.p $ maybe "" show rr.priority ]
        , HH.td_ [ Bulma.p rr.target ]
        , if rr.readonly
          then HH.td_ [ Bulma.btn_readonly ]
          else HH.td_ [ Bulma.btn_modify (CreateUpdateRRModal rr.rrid) ]
        , if rr.readonly
          then HH.td_ [ Bulma.btn_readonly ]
          else HH.td_ [ Bulma.btn_delete (DeleteRRModal rr.rrid) ]
        ]
      _     ->
        [ Bulma.txt_name rr.rrtype
        , HH.td_ [ Bulma.p rr.name]
        , HH.td_ [ Bulma.p $ show rr.ttl ]
        , HH.td_ [ Bulma.p rr.target ]
        ] <> if rr.readonly
          then [ HH.td_ [ Bulma.btn_readonly ] ]
          else [ HH.td_ [ Bulma.btn_modify (CreateUpdateRRModal rr.rrid) ]
               , HH.td_ [ Bulma.btn_delete (DeleteRRModal rr.rrid) ]
               ]

baseRecords :: Array String
baseRecords = [ "A", "AAAA", "CNAME", "TXT", "NS" ]

-- Component definition and initial state

render_new_records :: forall (w :: Type). State -> HH.HTML w Action
render_new_records _
  = Bulma.hdiv
      [ Bulma.h1 "Adding new records"
      -- use "level" to get horizontal buttons next to each other (probably vertical on mobile)
      , Bulma.level [
            Bulma.btn     "A" (CreateNewRRModal A)
          , Bulma.btn  "AAAA" (CreateNewRRModal AAAA)
          , Bulma.btn   "TXT" (CreateNewRRModal TXT)
          , Bulma.btn "CNAME" (CreateNewRRModal CNAME)
          , Bulma.btn    "NS" (CreateNewRRModal NS)
          , Bulma.btn    "MX" (CreateNewRRModal MX)
          , Bulma.btn   "SRV" (CreateNewRRModal SRV)
          ] []
      , Bulma.hr
      , Bulma.h1 "Special records about the mail system (soon)"
      -- use "level" to get horizontal buttons next to each other (probably vertical on mobile)
      , Bulma.level [
            Bulma.btn_ro (C.is_small <> C.is_warning) "SPF"
          , Bulma.btn_ro (C.is_small <> C.is_warning) "DKIM"
          , Bulma.btn_ro (C.is_small <> C.is_warning) "DMARC"
          ] []
      , Bulma.hr
      , Bulma.level [
            Bulma.btn "Get the final zone file." AskZoneFile
          ] [HH.text "For debug purposes. ⚠"]
      ]

render_zonefile :: forall (w :: Type). Maybe String -> HH.HTML w Action
render_zonefile zonefile = Bulma.box [ maybe (HH.text "") (\x -> HH.pre_ [HH.text x]) zonefile ]

-- ACTIONS

first :: forall a. (a -> Boolean) -> Array a -> Maybe a
first condition = A.head <<< (A.filter condition)

loopE :: forall state action input output m a b
  .  (a -> H.HalogenM state action input output m b)
  -> Array a
  -> H.HalogenM state action input output m Unit
loopE f a = case (A.head a) of
  Nothing -> pure unit
  Just x  -> do void $ f x
                case (A.tail a) of
                  Nothing -> pure unit
                  Just xs -> loopE f xs

update_field :: ResourceRecord -> Field -> ResourceRecord
update_field rr updated_field = case updated_field of
  Field_Domain val   -> rr { name = val }
  Field_Target val   -> rr { target = val }
  Field_TTL val      -> rr { ttl = fromMaybe 0 (fromString val) }
  Field_Priority val -> rr { priority = fromString val }
  Field_Protocol val -> rr { protocol = Just val }
  Field_Weight val   -> rr { weight = fromString val }
  Field_Port val     -> rr { port = fromString val }
